@@ -5,7 +5,6 @@
 package gc
 
 import (
-	"cmd/internal/obj"
 	"fmt"
 )
 
@@ -16,6 +15,15 @@ const (
 	InitPending    = 2
 )
 
+type InitEntry struct {
+	Xoffset int64 // struct, array only
+	Expr    *Node // bytes of run-time computed expressions
+}
+
+type InitPlan struct {
+	E []InitEntry
+}
+
 var (
 	initlist  []*Node
 	initplans map[*Node]*InitPlan
@@ -24,20 +32,20 @@ var (
 
 // init1 walks the AST starting at n, and accumulates in out
 // the list of definitions needing init code in dependency order.
-func init1(n *Node, out **NodeList) {
+func init1(n *Node, out *[]*Node) {
 	if n == nil {
 		return
 	}
 	init1(n.Left, out)
 	init1(n.Right, out)
-	for l := n.List; l != nil; l = l.Next {
-		init1(l.N, out)
+	for _, n1 := range n.List.Slice() {
+		init1(n1, out)
 	}
 
 	if n.Left != nil && n.Type != nil && n.Left.Op == OTYPE && n.Class == PFUNC {
 		// Methods called as Type.Method(receiver, ...).
 		// Definitions for method expressions are stored in type->nname.
-		init1(n.Type.Nname, out)
+		init1(n.Type.Nname(), out)
 	}
 
 	if n.Op != ONAME {
@@ -120,7 +128,7 @@ func init1(n *Node, out **NodeList) {
 				if Debug['%'] != 0 {
 					Dump("nonstatic", defn)
 				}
-				*out = list(*out, defn)
+				*out = append(*out, defn)
 			}
 
 		case OAS2FUNC, OAS2MAPR, OAS2DOTTYPE, OAS2RECV:
@@ -128,13 +136,13 @@ func init1(n *Node, out **NodeList) {
 				break
 			}
 			defn.Initorder = InitPending
-			for l := defn.Rlist; l != nil; l = l.Next {
-				init1(l.N, out)
+			for _, n2 := range defn.Rlist.Slice() {
+				init1(n2, out)
 			}
 			if Debug['%'] != 0 {
 				Dump("nonstatic", defn)
 			}
-			*out = list(*out, defn)
+			*out = append(*out, defn)
 			defn.Initorder = InitDone
 		}
 	}
@@ -187,13 +195,13 @@ func foundinitloop(node, visited *Node) {
 }
 
 // recurse over n, doing init1 everywhere.
-func init2(n *Node, out **NodeList) {
+func init2(n *Node, out *[]*Node) {
 	if n == nil || n.Initorder == InitDone {
 		return
 	}
 
-	if n.Op == ONAME && n.Ninit != nil {
-		Fatalf("name %v with ninit: %v\n", n.Sym, Nconv(n, obj.FmtSign))
+	if n.Op == ONAME && n.Ninit.Len() != 0 {
+		Fatalf("name %v with ninit: %v\n", n.Sym, Nconv(n, FmtSign))
 	}
 
 	init1(n, out)
@@ -208,28 +216,26 @@ func init2(n *Node, out **NodeList) {
 		init2list(n.Func.Closure.Nbody, out)
 	}
 	if n.Op == ODOTMETH || n.Op == OCALLPART {
-		init2(n.Type.Nname, out)
+		init2(n.Type.Nname(), out)
 	}
 }
 
-func init2list(l *NodeList, out **NodeList) {
-	for ; l != nil; l = l.Next {
-		init2(l.N, out)
+func init2list(l Nodes, out *[]*Node) {
+	for _, n := range l.Slice() {
+		init2(n, out)
 	}
 }
 
-func initreorder(l *NodeList, out **NodeList) {
+func initreorder(l []*Node, out *[]*Node) {
 	var n *Node
-
-	for ; l != nil; l = l.Next {
-		n = l.N
+	for _, n = range l {
 		switch n.Op {
 		case ODCLFUNC, ODCLCONST, ODCLTYPE:
 			continue
 		}
 
-		initreorder(n.Ninit, out)
-		n.Ninit = nil
+		initreorder(n.Ninit.Slice(), out)
+		n.Ninit.Set(nil)
 		init1(n, out)
 	}
 }
@@ -237,19 +243,19 @@ func initreorder(l *NodeList, out **NodeList) {
 // initfix computes initialization order for a list l of top-level
 // declarations and outputs the corresponding list of statements
 // to include in the init() function body.
-func initfix(l *NodeList) *NodeList {
-	var lout *NodeList
+func initfix(l []*Node) []*Node {
+	var lout []*Node
 	initplans = make(map[*Node]*InitPlan)
-	lno := int(lineno)
+	lno := lineno
 	initreorder(l, &lout)
-	lineno = int32(lno)
+	lineno = lno
 	initplans = nil
 	return lout
 }
 
 // compilation of top-level (static) assignments
 // into DATA statements if at all possible.
-func staticinit(n *Node, out **NodeList) bool {
+func staticinit(n *Node, out *[]*Node) bool {
 	if n.Op != ONAME || n.Class != PEXTERN || n.Name.Defn == nil || n.Name.Defn.Op != OAS {
 		Fatalf("staticinit")
 	}
@@ -262,7 +268,7 @@ func staticinit(n *Node, out **NodeList) bool {
 
 // like staticassign but we are copying an already
 // initialized value r.
-func staticcopy(l *Node, r *Node, out **NodeList) bool {
+func staticcopy(l *Node, r *Node, out *[]*Node) bool {
 	if r.Op != ONAME {
 		return false
 	}
@@ -291,7 +297,7 @@ func staticcopy(l *Node, r *Node, out **NodeList) bool {
 		if staticcopy(l, r, out) {
 			return true
 		}
-		*out = list(*out, Nod(OAS, l, r))
+		*out = append(*out, Nod(OAS, l, r))
 		return true
 
 	case OLITERAL:
@@ -322,7 +328,7 @@ func staticcopy(l *Node, r *Node, out **NodeList) bool {
 		}
 
 	case OARRAYLIT:
-		if Isslice(r.Type) {
+		if r.Type.IsSlice() {
 			// copy slice
 			a := inittemps[r]
 
@@ -336,8 +342,6 @@ func staticcopy(l *Node, r *Node, out **NodeList) bool {
 			return true
 		}
 		fallthrough
-
-		// fall through
 	case OSTRUCTLIT:
 		p := initplans[r]
 
@@ -362,7 +366,7 @@ func staticcopy(l *Node, r *Node, out **NodeList) bool {
 					rr.Type = ll.Type
 					rr.Xoffset += e.Xoffset
 					setlineno(rr)
-					*out = list(*out, Nod(OAS, ll, rr))
+					*out = append(*out, Nod(OAS, ll, rr))
 				}
 			}
 		}
@@ -373,7 +377,7 @@ func staticcopy(l *Node, r *Node, out **NodeList) bool {
 	return false
 }
 
-func staticassign(l *Node, r *Node, out **NodeList) bool {
+func staticassign(l *Node, r *Node, out *[]*Node) bool {
 	for r.Op == OCONVNOP {
 		r = r.Left
 	}
@@ -410,7 +414,7 @@ func staticassign(l *Node, r *Node, out **NodeList) bool {
 
 			// Init underlying literal.
 			if !staticassign(a, r.Left, out) {
-				*out = list(*out, Nod(OAS, a, r.Left))
+				*out = append(*out, Nod(OAS, a, r.Left))
 			}
 			return true
 		}
@@ -425,12 +429,10 @@ func staticassign(l *Node, r *Node, out **NodeList) bool {
 
 	case OARRAYLIT:
 		initplan(r)
-		if Isslice(r.Type) {
+		if r.Type.IsSlice() {
 			// Init slice.
-			ta := typ(TARRAY)
-
-			ta.Type = r.Type.Type
-			ta.Bound = Mpgetfix(r.Right.Val().U.(*Mpint))
+			bound := r.Right.Int64()
+			ta := typArray(r.Type.Elem(), bound)
 			a := staticname(ta, 1)
 			inittemps[r] = a
 			n := *l
@@ -463,7 +465,7 @@ func staticassign(l *Node, r *Node, out **NodeList) bool {
 				*a = n
 				a.Orig = a // completely separate copy
 				if !staticassign(a, e.Expr, out) {
-					*out = list(*out, Nod(OAS, a, e.Expr))
+					*out = append(*out, Nod(OAS, a, e.Expr))
 				}
 			}
 		}
@@ -475,12 +477,17 @@ func staticassign(l *Node, r *Node, out **NodeList) bool {
 		break
 
 	case OCLOSURE:
-		if r.Func.Cvars == nil {
+		if hasemptycvars(r) {
+			if Debug_closure > 0 {
+				Warnl(r.Lineno, "closure converted to global")
+			}
 			// Closures with no captured variables are globals,
 			// so the assignment can be done at link time.
 			n := *l
 			gdata(&n, r.Func.Closure.Func.Nname, Widthptr)
 			return true
+		} else {
+			closuredebugruntimecheck(r)
 		}
 	}
 
@@ -494,7 +501,7 @@ func staticassign(l *Node, r *Node, out **NodeList) bool {
 // data statements for the constant
 // part of the composite literal.
 func staticname(t *Type, ctxt int) *Node {
-	n := newname(Lookupf("statictmp_%.4d", statuniqgen))
+	n := newname(LookupN("statictmp_", statuniqgen))
 	statuniqgen++
 	if ctxt == 0 {
 		n.Name.Readonly = true
@@ -504,91 +511,101 @@ func staticname(t *Type, ctxt int) *Node {
 }
 
 func isliteral(n *Node) bool {
-	if n.Op == OLITERAL {
-		if n.Val().Ctype() != CTNIL {
-			return true
-		}
-	}
-	return false
+	// Treat nils as zeros rather than literals.
+	return n.Op == OLITERAL && n.Val().Ctype() != CTNIL
 }
 
-func simplename(n *Node) bool {
-	if n.Op != ONAME {
+func (n *Node) isSimpleName() bool {
+	return n.Op == ONAME && n.Addable && n.Class != PAUTOHEAP
+}
+
+func litas(l *Node, r *Node, init *Nodes) {
+	a := Nod(OAS, l, r)
+	a = typecheck(a, Etop)
+	a = walkexpr(a, init)
+	init.Append(a)
+}
+
+// initGenType is a bitmap indicating the types of generation that will occur for a static value.
+type initGenType uint8
+
+const (
+	initDynamic initGenType = 1 << iota // contains some dynamic values, for which init code will be generated
+	initConst                           // contains some constant values, which may be written into data symbols
+)
+
+func getdyn(n *Node, top int) initGenType {
+	switch n.Op {
+	default:
+		if isliteral(n) {
+			return initConst
+		}
+		return initDynamic
+
+	case OARRAYLIT:
+		if top == 0 && n.Type.IsSlice() {
+			return initDynamic
+		}
+
+	case OSTRUCTLIT:
+	}
+
+	var mode initGenType
+	for _, n1 := range n.List.Slice() {
+		value := n1.Right
+		mode |= getdyn(value, 0)
+		if mode == initDynamic|initConst {
+			break
+		}
+	}
+	return mode
+}
+
+// isStaticCompositeLiteral reports whether n is a compile-time constant.
+func isStaticCompositeLiteral(n *Node) bool {
+	switch n.Op {
+	case OARRAYLIT:
+		if n.Type.IsSlice() {
+			return false
+		}
+	case OSTRUCTLIT:
+	case OLITERAL:
+		return true
+	default:
 		return false
 	}
-	if !n.Addable {
-		return false
-	}
-	if n.Class&PHEAP != 0 {
-		return false
-	}
-	if n.Class == PPARAMREF {
-		return false
+	for _, r := range n.List.Slice() {
+		if r.Op != OKEY {
+			Fatalf("isStaticCompositeLiteral: rhs not OKEY: %v", r)
+		}
+		index := r.Left
+		if n.Op == OARRAYLIT && index.Op != OLITERAL {
+			return false
+		}
+		value := r.Right
+		if !isStaticCompositeLiteral(value) {
+			return false
+		}
 	}
 	return true
 }
 
-func litas(l *Node, r *Node, init **NodeList) {
-	a := Nod(OAS, l, r)
-	typecheck(&a, Etop)
-	walkexpr(&a, init)
-	*init = list(*init, a)
-}
-
-const (
-	MODEDYNAM = 1
-	MODECONST = 2
-)
-
-func getdyn(n *Node, top int) int {
-	mode := 0
-	switch n.Op {
-	default:
-		if isliteral(n) {
-			return MODECONST
-		}
-		return MODEDYNAM
-
-	case OARRAYLIT:
-		if top == 0 && n.Type.Bound < 0 {
-			return MODEDYNAM
-		}
-		fallthrough
-
-	case OSTRUCTLIT:
-		break
-	}
-
-	for nl := n.List; nl != nil; nl = nl.Next {
-		value := nl.N.Right
-		mode |= getdyn(value, 0)
-		if mode == MODEDYNAM|MODECONST {
-			break
-		}
-	}
-
-	return mode
-}
-
-func structlit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
-	for nl := n.List; nl != nil; nl = nl.Next {
-		r := nl.N
+func structlit(ctxt int, pass int, n *Node, var_ *Node, init *Nodes) {
+	for _, r := range n.List.Slice() {
 		if r.Op != OKEY {
 			Fatalf("structlit: rhs not OKEY: %v", r)
 		}
 		index := r.Left
 		value := r.Right
 
-		var a *Node
-
 		switch value.Op {
 		case OARRAYLIT:
-			if value.Type.Bound < 0 {
+			if value.Type.IsSlice() {
 				if pass == 1 && ctxt != 0 {
-					a = Nod(ODOT, var_, newname(index.Sym))
+					a := NodSym(ODOT, var_, index.Sym)
 					slicelit(ctxt, value, a, init)
 				} else if pass == 2 && ctxt == 0 {
-					a = Nod(ODOT, var_, newname(index.Sym))
+					a := NodSym(ODOT, var_, index.Sym)
 					slicelit(ctxt, value, a, init)
 				} else if pass == 3 {
 					break
@@ -596,12 +613,12 @@ func structlit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
 				continue
 			}
 
-			a = Nod(ODOT, var_, newname(index.Sym))
+			a := NodSym(ODOT, var_, index.Sym)
 			arraylit(ctxt, pass, value, a, init)
 			continue
 
 		case OSTRUCTLIT:
-			a = Nod(ODOT, var_, newname(index.Sym))
+			a := NodSym(ODOT, var_, index.Sym)
 			structlit(ctxt, pass, value, a, init)
 			continue
 		}
@@ -616,44 +633,41 @@ func structlit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
 
 		// build list of var.field = expr
 		setlineno(value)
-		a = Nod(ODOT, var_, newname(index.Sym))
+		a := NodSym(ODOT, var_, index.Sym)
 
 		a = Nod(OAS, a, value)
-		typecheck(&a, Etop)
+		a = typecheck(a, Etop)
 		if pass == 1 {
-			walkexpr(&a, init) // add any assignments in r to top
+			a = walkexpr(a, init) // add any assignments in r to top
 			if a.Op != OAS {
 				Fatalf("structlit: not as")
 			}
 			a.Dodata = 2
 		} else {
-			orderstmtinplace(&a)
-			walkstmt(&a)
+			a = orderstmtinplace(a)
+			a = walkstmt(a)
 		}
 
-		*init = list(*init, a)
+		init.Append(a)
 	}
 }
 
-func arraylit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
-	for l := n.List; l != nil; l = l.Next {
-		r := l.N
+func arraylit(ctxt int, pass int, n *Node, var_ *Node, init *Nodes) {
+	for _, r := range n.List.Slice() {
 		if r.Op != OKEY {
 			Fatalf("arraylit: rhs not OKEY: %v", r)
 		}
 		index := r.Left
 		value := r.Right
 
-		var a *Node
-
 		switch value.Op {
 		case OARRAYLIT:
-			if value.Type.Bound < 0 {
+			if value.Type.IsSlice() {
 				if pass == 1 && ctxt != 0 {
-					a = Nod(OINDEX, var_, index)
+					a := Nod(OINDEX, var_, index)
 					slicelit(ctxt, value, a, init)
 				} else if pass == 2 && ctxt == 0 {
-					a = Nod(OINDEX, var_, index)
+					a := Nod(OINDEX, var_, index)
 					slicelit(ctxt, value, a, init)
 				} else if pass == 3 {
 					break
@@ -661,12 +675,12 @@ func arraylit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
 				continue
 			}
 
-			a = Nod(OINDEX, var_, index)
+			a := Nod(OINDEX, var_, index)
 			arraylit(ctxt, pass, value, a, init)
 			continue
 
 		case OSTRUCTLIT:
-			a = Nod(OINDEX, var_, index)
+			a := Nod(OINDEX, var_, index)
 			structlit(ctxt, pass, value, a, init)
 			continue
 		}
@@ -681,33 +695,28 @@ func arraylit(ctxt int, pass int, n *Node, var_ *Node, init **NodeList) {
 
 		// build list of var[index] = value
 		setlineno(value)
-		a = Nod(OINDEX, var_, index)
+		a := Nod(OINDEX, var_, index)
 
 		a = Nod(OAS, a, value)
-		typecheck(&a, Etop)
+		a = typecheck(a, Etop)
 		if pass == 1 {
-			walkexpr(&a, init)
+			a = walkexpr(a, init)
 			if a.Op != OAS {
 				Fatalf("arraylit: not as")
 			}
 			a.Dodata = 2
 		} else {
-			orderstmtinplace(&a)
-			walkstmt(&a)
+			a = orderstmtinplace(a)
+			a = walkstmt(a)
 		}
 
-		*init = list(*init, a)
+		init.Append(a)
 	}
 }
 
-func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
-	// make an array type
-	t := shallow(n.Type)
-
-	t.Bound = Mpgetfix(n.Right.Val().U.(*Mpint))
-	t.Width = 0
-	t.Sym = nil
-	t.Haspointers = 0
+func slicelit(ctxt int, n *Node, var_ *Node, init *Nodes) {
+	// make an array type corresponding the number of elements we have
+	t := typArray(n.Type.Elem(), n.Right.Int64())
 	dowidth(t)
 
 	if ctxt != 0 {
@@ -718,12 +727,12 @@ func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 		arraylit(ctxt, 2, n, vstat, init)
 
 		// copy static to slice
-		a := Nod(OSLICE, vstat, Nod(OKEY, nil, nil))
+		a := Nod(OSLICE, vstat, nil)
 
 		a = Nod(OAS, var_, a)
-		typecheck(&a, Etop)
+		a = typecheck(a, Etop)
 		a.Dodata = 2
-		*init = list(*init, a)
+		init.Append(a)
 		return
 	}
 
@@ -736,22 +745,22 @@ func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 	//	var vauto *[...]t = new([...]t)
 	// 4. copy the static array to the auto array
 	//	*vauto = vstat
-	// 5. assign slice of allocated heap to var
-	//	var = [0:]*auto
-	// 6. for each dynamic part assign to the slice
-	//	var[i] = dynamic part
+	// 5. for each dynamic part assign to the array
+	//	vauto[i] = dynamic part
+	// 6. assign slice of allocated heap to var
+	//	var = vauto[:]
 	//
 	// an optimization is done if there is no constant part
 	//	3. var vauto *[...]t = new([...]t)
-	//	5. var = [0:]*auto
-	//	6. var[i] = dynamic part
+	//	5. vauto[i] = dynamic part
+	//	6. var = vauto[:]
 
 	// if the literal contains constants,
 	// make static initialized array (1),(2)
 	var vstat *Node
 
 	mode := getdyn(n, 1)
-	if mode&MODECONST != 0 {
+	if mode&initConst != 0 {
 		vstat = staticname(t, ctxt)
 		arraylit(ctxt, 1, n, vstat, init)
 	}
@@ -767,8 +776,8 @@ func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 
 		if vstat == nil {
 			a = Nod(OAS, x, nil)
-			typecheck(&a, Etop)
-			*init = list(*init, a) // zero new temp
+			a = typecheck(a, Etop)
+			init.Append(a) // zero new temp
 		}
 
 		a = Nod(OADDR, x, nil)
@@ -776,56 +785,47 @@ func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 		a = temp(t)
 		if vstat == nil {
 			a = Nod(OAS, temp(t), nil)
-			typecheck(&a, Etop)
-			*init = list(*init, a) // zero new temp
+			a = typecheck(a, Etop)
+			init.Append(a) // zero new temp
 			a = a.Left
 		}
 
 		a = Nod(OADDR, a, nil)
 	} else {
 		a = Nod(ONEW, nil, nil)
-		a.List = list1(typenod(t))
+		a.List.Set1(typenod(t))
 	}
 
 	a = Nod(OAS, vauto, a)
-	typecheck(&a, Etop)
-	walkexpr(&a, init)
-	*init = list(*init, a)
+	a = typecheck(a, Etop)
+	a = walkexpr(a, init)
+	init.Append(a)
 
 	if vstat != nil {
 		// copy static to heap (4)
 		a = Nod(OIND, vauto, nil)
 
 		a = Nod(OAS, a, vstat)
-		typecheck(&a, Etop)
-		walkexpr(&a, init)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = walkexpr(a, init)
+		init.Append(a)
 	}
 
-	// make slice out of heap (5)
-	a = Nod(OAS, var_, Nod(OSLICE, vauto, Nod(OKEY, nil, nil)))
-
-	typecheck(&a, Etop)
-	orderstmtinplace(&a)
-	walkstmt(&a)
-	*init = list(*init, a)
-
-	// put dynamics into slice (6)
-	for l := n.List; l != nil; l = l.Next {
-		r := l.N
+	// put dynamics into array (5)
+	for _, r := range n.List.Slice() {
 		if r.Op != OKEY {
 			Fatalf("slicelit: rhs not OKEY: %v", r)
 		}
 		index := r.Left
 		value := r.Right
-		a := Nod(OINDEX, var_, index)
+		a := Nod(OINDEX, vauto, index)
 		a.Bounded = true
 
 		// TODO need to check bounds?
 
 		switch value.Op {
 		case OARRAYLIT:
-			if value.Type.Bound < 0 {
+			if value.Type.IsSlice() {
 				break
 			}
 			arraylit(ctxt, 2, value, a, init)
@@ -840,32 +840,38 @@ func slicelit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 			continue
 		}
 
-		// build list of var[c] = expr
+		// build list of vauto[c] = expr
 		setlineno(value)
 		a = Nod(OAS, a, value)
 
-		typecheck(&a, Etop)
-		orderstmtinplace(&a)
-		walkstmt(&a)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = orderstmtinplace(a)
+		a = walkstmt(a)
+		init.Append(a)
 	}
+
+	// make slice out of heap (6)
+	a = Nod(OAS, var_, Nod(OSLICE, vauto, nil))
+
+	a = typecheck(a, Etop)
+	a = orderstmtinplace(a)
+	a = walkstmt(a)
+	init.Append(a)
 }
 
-func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
+func maplit(ctxt int, n *Node, var_ *Node, init *Nodes) {
 	ctxt = 0
 
 	// make the map var
 	nerr := nerrors
 
 	a := Nod(OMAKE, nil, nil)
-	a.List = list1(typenod(n.Type))
+	a.List.Set1(typenod(n.Type))
 	litas(var_, a, init)
 
 	// count the initializers
-	b := int64(0)
-
-	for l := n.List; l != nil; l = l.Next {
-		r := l.N
+	b := 0
+	for _, r := range n.List.Slice() {
 		if r.Op != OKEY {
 			Fatalf("maplit: rhs not OKEY: %v", r)
 		}
@@ -880,40 +886,33 @@ func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 	if b != 0 {
 		// build type [count]struct { a Tindex, b Tvalue }
 		t := n.Type
-
-		tk := t.Down
-		tv := t.Type
-
-		symb := Lookup("b")
-		t = typ(TFIELD)
-		t.Type = tv
-		t.Sym = symb
+		tk := t.Key()
+		tv := t.Val()
 
 		syma := Lookup("a")
-		t1 := t
-		t = typ(TFIELD)
-		t.Type = tk
-		t.Sym = syma
-		t.Down = t1
+		symb := Lookup("b")
 
-		t1 = t
-		t = typ(TSTRUCT)
-		t.Type = t1
+		var fields [2]*Field
+		fields[0] = newField()
+		fields[0].Type = tk
+		fields[0].Sym = syma
+		fields[1] = newField()
+		fields[1].Type = tv
+		fields[1].Sym = symb
 
-		t1 = t
-		t = typ(TARRAY)
-		t.Bound = b
-		t.Type = t1
+		tstruct := typ(TSTRUCT)
+		tstruct.SetFields(fields[:])
 
-		dowidth(t)
+		tarr := typArray(tstruct, int64(b))
+
+		// TODO(josharian): suppress alg generation for these types?
+		dowidth(tarr)
 
 		// make and initialize static array
-		vstat := staticname(t, ctxt)
+		vstat := staticname(tarr, ctxt)
 
 		b := int64(0)
-		for l := n.List; l != nil; l = l.Next {
-			r := l.N
-
+		for _, r := range n.List.Slice() {
 			if r.Op != OKEY {
 				Fatalf("maplit: rhs not OKEY: %v", r)
 			}
@@ -926,24 +925,24 @@ func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 				a = Nodintconst(b)
 
 				a = Nod(OINDEX, vstat, a)
-				a = Nod(ODOT, a, newname(syma))
+				a = NodSym(ODOT, a, syma)
 				a = Nod(OAS, a, index)
-				typecheck(&a, Etop)
-				walkexpr(&a, init)
+				a = typecheck(a, Etop)
+				a = walkexpr(a, init)
 				a.Dodata = 2
-				*init = list(*init, a)
+				init.Append(a)
 
 				// build vstat[b].b = value;
 				setlineno(value)
 				a = Nodintconst(b)
 
 				a = Nod(OINDEX, vstat, a)
-				a = Nod(ODOT, a, newname(symb))
+				a = NodSym(ODOT, a, symb)
 				a = Nod(OAS, a, value)
-				typecheck(&a, Etop)
-				walkexpr(&a, init)
+				a = typecheck(a, Etop)
+				a = walkexpr(a, init)
 				a.Dodata = 2
-				*init = list(*init, a)
+				init.Append(a)
 
 				b++
 			}
@@ -957,34 +956,30 @@ func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 
 		a = Nod(OINDEX, vstat, index)
 		a.Bounded = true
-		a = Nod(ODOT, a, newname(symb))
+		a = NodSym(ODOT, a, symb)
 
 		r := Nod(OINDEX, vstat, index)
 		r.Bounded = true
-		r = Nod(ODOT, r, newname(syma))
+		r = NodSym(ODOT, r, syma)
 		r = Nod(OINDEX, var_, r)
 
 		r = Nod(OAS, r, a)
 
 		a = Nod(OFOR, nil, nil)
-		a.Nbody = list1(r)
+		a.Nbody.Set1(r)
 
-		a.Ninit = list1(Nod(OAS, index, Nodintconst(0)))
-		a.Left = Nod(OLT, index, Nodintconst(t.Bound))
+		a.Ninit.Set1(Nod(OAS, index, Nodintconst(0)))
+		a.Left = Nod(OLT, index, Nodintconst(tarr.NumElem()))
 		a.Right = Nod(OAS, index, Nod(OADD, index, Nodintconst(1)))
 
-		typecheck(&a, Etop)
-		walkstmt(&a)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = walkstmt(a)
+		init.Append(a)
 	}
 
 	// put in dynamic entries one-at-a-time
-	var key *Node
-
-	var val *Node
-	for l := n.List; l != nil; l = l.Next {
-		r := l.N
-
+	var key, val *Node
+	for _, r := range n.List.Slice() {
 		if r.Op != OKEY {
 			Fatalf("maplit: rhs not OKEY: %v", r)
 		}
@@ -998,26 +993,26 @@ func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 		// build list of var[c] = expr.
 		// use temporary so that mapassign1 can have addressable key, val.
 		if key == nil {
-			key = temp(var_.Type.Down)
-			val = temp(var_.Type.Type)
+			key = temp(var_.Type.Key())
+			val = temp(var_.Type.Val())
 		}
 
 		setlineno(r.Left)
 		a = Nod(OAS, key, r.Left)
-		typecheck(&a, Etop)
-		walkstmt(&a)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = walkstmt(a)
+		init.Append(a)
 		setlineno(r.Right)
 		a = Nod(OAS, val, r.Right)
-		typecheck(&a, Etop)
-		walkstmt(&a)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = walkstmt(a)
+		init.Append(a)
 
 		setlineno(val)
 		a = Nod(OAS, Nod(OINDEX, var_, key), val)
-		typecheck(&a, Etop)
-		walkstmt(&a)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		a = walkstmt(a)
+		init.Append(a)
 
 		if nerr != nerrors {
 			break
@@ -1026,29 +1021,29 @@ func maplit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 
 	if key != nil {
 		a = Nod(OVARKILL, key, nil)
-		typecheck(&a, Etop)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		init.Append(a)
 		a = Nod(OVARKILL, val, nil)
-		typecheck(&a, Etop)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		init.Append(a)
 	}
 }
 
-func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
+func anylit(ctxt int, n *Node, var_ *Node, init *Nodes) {
 	t := n.Type
 	switch n.Op {
 	default:
-		Fatalf("anylit: not lit")
+		Fatalf("anylit: not lit, op=%v node=%v", n.Op, n)
 
 	case OPTRLIT:
-		if !Isptr[t.Etype] {
+		if !t.IsPtr() {
 			Fatalf("anylit: not ptr")
 		}
 
 		var r *Node
 		if n.Right != nil {
 			r = Nod(OADDR, n.Right, nil)
-			typecheck(&r, Erv)
+			r = typecheck(r, Erv)
 		} else {
 			r = Nod(ONEW, nil, nil)
 			r.Typecheck = 1
@@ -1056,22 +1051,22 @@ func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 			r.Esc = n.Esc
 		}
 
-		walkexpr(&r, init)
+		r = walkexpr(r, init)
 		a := Nod(OAS, var_, r)
 
-		typecheck(&a, Etop)
-		*init = list(*init, a)
+		a = typecheck(a, Etop)
+		init.Append(a)
 
 		var_ = Nod(OIND, var_, nil)
-		typecheck(&var_, Erv|Easgn)
+		var_ = typecheck(var_, Erv|Easgn)
 		anylit(ctxt, n.Left, var_, init)
 
 	case OSTRUCTLIT:
-		if t.Etype != TSTRUCT {
+		if !t.IsStruct() {
 			Fatalf("anylit: not struct")
 		}
 
-		if simplename(var_) && count(n.List) > 4 {
+		if var_.isSimpleName() && n.List.Len() > 4 {
 			if ctxt == 0 {
 				// lay out static data
 				vstat := staticname(t, ctxt)
@@ -1081,9 +1076,9 @@ func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 				// copy static to var
 				a := Nod(OAS, var_, vstat)
 
-				typecheck(&a, Etop)
-				walkexpr(&a, init)
-				*init = list(*init, a)
+				a = typecheck(a, Etop)
+				a = walkexpr(a, init)
+				init.Append(a)
 
 				// add expressions to automatic
 				structlit(ctxt, 2, n, var_, init)
@@ -1097,25 +1092,25 @@ func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 		}
 
 		// initialize of not completely specified
-		if simplename(var_) || count(n.List) < structcount(t) {
+		if var_.isSimpleName() || n.List.Len() < t.NumFields() {
 			a := Nod(OAS, var_, nil)
-			typecheck(&a, Etop)
-			walkexpr(&a, init)
-			*init = list(*init, a)
+			a = typecheck(a, Etop)
+			a = walkexpr(a, init)
+			init.Append(a)
 		}
 
 		structlit(ctxt, 3, n, var_, init)
 
 	case OARRAYLIT:
-		if t.Etype != TARRAY {
-			Fatalf("anylit: not array")
-		}
-		if t.Bound < 0 {
+		if t.IsSlice() {
 			slicelit(ctxt, n, var_, init)
 			break
 		}
+		if !t.IsArray() {
+			Fatalf("anylit: not array")
+		}
 
-		if simplename(var_) && count(n.List) > 4 {
+		if var_.isSimpleName() && n.List.Len() > 4 {
 			if ctxt == 0 {
 				// lay out static data
 				vstat := staticname(t, ctxt)
@@ -1125,9 +1120,9 @@ func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 				// copy static to automatic
 				a := Nod(OAS, var_, vstat)
 
-				typecheck(&a, Etop)
-				walkexpr(&a, init)
-				*init = list(*init, a)
+				a = typecheck(a, Etop)
+				a = walkexpr(a, init)
+				init.Append(a)
 
 				// add expressions to automatic
 				arraylit(ctxt, 2, n, var_, init)
@@ -1141,24 +1136,24 @@ func anylit(ctxt int, n *Node, var_ *Node, init **NodeList) {
 		}
 
 		// initialize of not completely specified
-		if simplename(var_) || int64(count(n.List)) < t.Bound {
+		if var_.isSimpleName() || int64(n.List.Len()) < t.NumElem() {
 			a := Nod(OAS, var_, nil)
-			typecheck(&a, Etop)
-			walkexpr(&a, init)
-			*init = list(*init, a)
+			a = typecheck(a, Etop)
+			a = walkexpr(a, init)
+			init.Append(a)
 		}
 
 		arraylit(ctxt, 3, n, var_, init)
 
 	case OMAPLIT:
-		if t.Etype != TMAP {
+		if !t.IsMap() {
 			Fatalf("anylit: not map")
 		}
 		maplit(ctxt, n, var_, init)
 	}
 }
 
-func oaslit(n *Node, init **NodeList) bool {
+func oaslit(n *Node, init *Nodes) bool {
 	if n.Left == nil || n.Right == nil {
 		// not a special composit literal assignment
 		return false
@@ -1167,7 +1162,7 @@ func oaslit(n *Node, init **NodeList) bool {
 		// not a special composit literal assignment
 		return false
 	}
-	if !simplename(n.Left) {
+	if !n.Left.isSimpleName() {
 		// not a special composit literal assignment
 		return false
 	}
@@ -1204,11 +1199,12 @@ func oaslit(n *Node, init **NodeList) bool {
 
 func getlit(lit *Node) int {
 	if Smallintconst(lit) {
-		return int(Mpgetfix(lit.Val().U.(*Mpint)))
+		return int(lit.Int64())
 	}
 	return -1
 }
 
+// stataddr sets nam to the static address of n and reports whether it succeeeded.
 func stataddr(nam *Node, n *Node) bool {
 	if n == nil {
 		return false
@@ -1228,7 +1224,7 @@ func stataddr(nam *Node, n *Node) bool {
 		return true
 
 	case OINDEX:
-		if n.Left.Type.Bound < 0 {
+		if n.Left.Type.IsSlice() {
 			break
 		}
 		if !stataddr(nam, n.Left) {
@@ -1262,38 +1258,34 @@ func initplan(n *Node) {
 		Fatalf("initplan")
 
 	case OARRAYLIT:
-		for l := n.List; l != nil; l = l.Next {
-			a := l.N
+		for _, a := range n.List.Slice() {
 			if a.Op != OKEY || !Smallintconst(a.Left) {
 				Fatalf("initplan arraylit")
 			}
-			addvalue(p, n.Type.Type.Width*Mpgetfix(a.Left.Val().U.(*Mpint)), nil, a.Right)
+			addvalue(p, n.Type.Elem().Width*a.Left.Int64(), a.Right)
 		}
 
 	case OSTRUCTLIT:
-		for l := n.List; l != nil; l = l.Next {
-			a := l.N
-			if a.Op != OKEY || a.Left.Type == nil {
+		for _, a := range n.List.Slice() {
+			if a.Op != OKEY || a.Left.Type != structkey {
 				Fatalf("initplan structlit")
 			}
-			addvalue(p, a.Left.Type.Width, nil, a.Right)
+			addvalue(p, a.Left.Xoffset, a.Right)
 		}
 
 	case OMAPLIT:
-		for l := n.List; l != nil; l = l.Next {
-			a := l.N
+		for _, a := range n.List.Slice() {
 			if a.Op != OKEY {
 				Fatalf("initplan maplit")
 			}
-			addvalue(p, -1, a.Left, a.Right)
+			addvalue(p, -1, a.Right)
 		}
 	}
 }
 
-func addvalue(p *InitPlan, xoffset int64, key *Node, n *Node) {
+func addvalue(p *InitPlan, xoffset int64, n *Node) {
 	// special case: zero can be dropped entirely
 	if iszero(n) {
-		p.Zero += n.Type.Width
 		return
 	}
 
@@ -1302,62 +1294,46 @@ func addvalue(p *InitPlan, xoffset int64, key *Node, n *Node) {
 		initplan(n)
 		q := initplans[n]
 		for _, qe := range q.E {
-			e := entry(p)
-			*e = qe
-			e.Xoffset += xoffset
+			// qe is a copy; we are not modifying entries in q.E
+			qe.Xoffset += xoffset
+			p.E = append(p.E, qe)
 		}
 		return
 	}
 
 	// add to plan
-	if n.Op == OLITERAL {
-		p.Lit += n.Type.Width
-	} else {
-		p.Expr += n.Type.Width
-	}
-
-	e := entry(p)
-	e.Xoffset = xoffset
-	e.Expr = n
+	p.E = append(p.E, InitEntry{Xoffset: xoffset, Expr: n})
 }
 
 func iszero(n *Node) bool {
 	switch n.Op {
 	case OLITERAL:
-		switch n.Val().Ctype() {
+		switch u := n.Val().U.(type) {
 		default:
 			Dump("unexpected literal", n)
 			Fatalf("iszero")
-
-		case CTNIL:
+		case *NilVal:
 			return true
-
-		case CTSTR:
-			return n.Val().U.(string) == ""
-
-		case CTBOOL:
-			return !n.Val().U.(bool)
-
-		case CTINT, CTRUNE:
-			return mpcmpfixc(n.Val().U.(*Mpint), 0) == 0
-
-		case CTFLT:
-			return mpcmpfltc(n.Val().U.(*Mpflt), 0) == 0
-
-		case CTCPLX:
-			return mpcmpfltc(&n.Val().U.(*Mpcplx).Real, 0) == 0 && mpcmpfltc(&n.Val().U.(*Mpcplx).Imag, 0) == 0
+		case string:
+			return u == ""
+		case bool:
+			return !u
+		case *Mpint:
+			return u.CmpInt64(0) == 0
+		case *Mpflt:
+			return u.CmpFloat64(0) == 0
+		case *Mpcplx:
+			return u.Real.CmpFloat64(0) == 0 && u.Imag.CmpFloat64(0) == 0
 		}
 
 	case OARRAYLIT:
-		if Isslice(n.Type) {
+		if n.Type.IsSlice() {
 			break
 		}
 		fallthrough
-
-		// fall through
 	case OSTRUCTLIT:
-		for l := n.List; l != nil; l = l.Next {
-			if !iszero(l.N.Right) {
+		for _, n1 := range n.List.Slice() {
+			if !iszero(n1.Right) {
 				return false
 			}
 		}
@@ -1368,93 +1344,86 @@ func iszero(n *Node) bool {
 }
 
 func isvaluelit(n *Node) bool {
-	return (n.Op == OARRAYLIT && Isfixedarray(n.Type)) || n.Op == OSTRUCTLIT
+	return (n.Op == OARRAYLIT && n.Type.IsArray()) || n.Op == OSTRUCTLIT
 }
 
-func entry(p *InitPlan) *InitEntry {
-	p.E = append(p.E, InitEntry{})
-	return &p.E[len(p.E)-1]
+// gen_as_init attempts to emit static data for n and reports whether it succeeded.
+// If reportOnly is true, it does not emit static data and does not modify the AST.
+func gen_as_init(n *Node, reportOnly bool) bool {
+	success := genAsInitNoCheck(n, reportOnly)
+	if !success && n.Dodata == 2 {
+		Dump("\ngen_as_init", n)
+		Fatalf("gen_as_init couldn't make data statement")
+	}
+	return success
 }
 
-func gen_as_init(n *Node) bool {
-	var nr *Node
-	var nl *Node
-	var nam Node
-
+func genAsInitNoCheck(n *Node, reportOnly bool) bool {
 	if n.Dodata == 0 {
-		goto no
+		return false
 	}
 
-	nr = n.Right
-	nl = n.Left
+	nr := n.Right
+	nl := n.Left
 	if nr == nil {
 		var nam Node
-		if !stataddr(&nam, nl) {
-			goto no
-		}
-		if nam.Class != PEXTERN {
-			goto no
-		}
-		return true
+		return stataddr(&nam, nl) && nam.Class == PEXTERN
 	}
 
 	if nr.Type == nil || !Eqtype(nl.Type, nr.Type) {
-		goto no
+		return false
 	}
 
-	if !stataddr(&nam, nl) {
-		goto no
-	}
-
-	if nam.Class != PEXTERN {
-		goto no
+	var nam Node
+	if !stataddr(&nam, nl) || nam.Class != PEXTERN {
+		return false
 	}
 
 	switch nr.Op {
 	default:
-		goto no
+		return false
 
 	case OCONVNOP:
 		nr = nr.Left
 		if nr == nil || nr.Op != OSLICEARR {
-			goto no
+			return false
 		}
 		fallthrough
 
-		// fall through
 	case OSLICEARR:
-		if nr.Right.Op == OKEY && nr.Right.Left == nil && nr.Right.Right == nil {
-			nr = nr.Left
-			gused(nil) // in case the data is the dest of a goto
-			nl := nr
-			if nr == nil || nr.Op != OADDR {
-				goto no
-			}
-			nr = nr.Left
-			if nr == nil || nr.Op != ONAME {
-				goto no
-			}
+		low, high, _ := nr.SliceBounds()
+		if low != nil || high != nil {
+			return false
+		}
+		nr = nr.Left
+		if nr == nil || nr.Op != OADDR {
+			return false
+		}
+		ptr := nr
+		nr = nr.Left
+		if nr == nil || nr.Op != ONAME {
+			return false
+		}
 
-			// nr is the array being converted to a slice
-			if nr.Type == nil || nr.Type.Etype != TARRAY || nr.Type.Bound < 0 {
-				goto no
-			}
+		// nr is the array being converted to a slice
+		if nr.Type == nil || !nr.Type.IsArray() {
+			return false
+		}
 
+		if !reportOnly {
 			nam.Xoffset += int64(Array_array)
-			gdata(&nam, nl, int(Types[Tptr].Width))
+			gdata(&nam, ptr, Widthptr)
 
 			nam.Xoffset += int64(Array_nel) - int64(Array_array)
 			var nod1 Node
-			Nodconst(&nod1, Types[TINT], nr.Type.Bound)
+			Nodconst(&nod1, Types[TINT], nr.Type.NumElem())
 			gdata(&nam, &nod1, Widthint)
 
 			nam.Xoffset += int64(Array_cap) - int64(Array_nel)
 			gdata(&nam, &nod1, Widthint)
-
-			return true
 		}
 
-		goto no
+		return true
 
 	case OLITERAL:
 		break
@@ -1462,40 +1431,27 @@ func gen_as_init(n *Node) bool {
 
 	switch nr.Type.Etype {
 	default:
-		goto no
+		return false
 
-	case TBOOL,
-		TINT8,
-		TUINT8,
-		TINT16,
-		TUINT16,
-		TINT32,
-		TUINT32,
-		TINT64,
-		TUINT64,
-		TINT,
-		TUINT,
-		TUINTPTR,
-		TPTR32,
-		TPTR64,
-		TFLOAT32,
-		TFLOAT64:
-		gdata(&nam, nr, int(nr.Type.Width))
+	case TBOOL, TINT8, TUINT8, TINT16, TUINT16,
+		TINT32, TUINT32, TINT64, TUINT64,
+		TINT, TUINT, TUINTPTR,
+		TPTR32, TPTR64,
+		TFLOAT32, TFLOAT64:
+		if !reportOnly {
+			gdata(&nam, nr, int(nr.Type.Width))
+		}
 
 	case TCOMPLEX64, TCOMPLEX128:
-		gdatacomplex(&nam, nr.Val().U.(*Mpcplx))
+		if !reportOnly {
+			gdatacomplex(&nam, nr.Val().U.(*Mpcplx))
+		}
 
 	case TSTRING:
-		gdatastring(&nam, nr.Val().U.(string))
+		if !reportOnly {
+			gdatastring(&nam, nr.Val().U.(string))
+		}
 	}
 
 	return true
-
-no:
-	if n.Dodata == 2 {
-		Dump("\ngen_as_init", n)
-		Fatalf("gen_as_init couldnt make data statement")
-	}
-
-	return false
 }
